// network.cpp - v2.5
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "stdafx.h"
#ifndef NO_SSL
#include <OpenSSL/err.h>
#include <OpenSSL/engine.h>
#include <OpenSSL/conf.h>
#endif //NO_SSL
#include "network.h"
#include "serialization.h"

namespace lib
{
	packet::packet() : iId(0)
	{
	}

	packet::packet(const packet& aPacket) : iId(aPacket.iId), iData(aPacket.iData)
	{
	}

	packet::packet(int aId) : iId(aId)
	{
	}

	packet::packet(int aId, size_t aLength, const void* aData) : iId(aId), iData(aLength)
	{
		if (aData)
			std::copy(static_cast<const char*>(aData), static_cast<const char*>(aData) + aLength, iData.begin());
	}

	packet::packet(int aId, char aData) : iId(aId), iData(sizeof(aData))
	{
		std::copy(&aData, &aData + sizeof(aData), iData.begin());
	}

	packet::packet(int aId, int aData) : iId(aId), iData(sizeof(aData))
	{
		std::copy(reinterpret_cast<const char*>(&aData), reinterpret_cast<const char*>(&aData) + sizeof(aData), iData.begin());
	}

	packet::packet(size_t aLength, const void* aData) : iId(0), iData(aLength)
	{
		if (aData)
			std::copy(static_cast<const char*>(aData), static_cast<const char*>(aData) + aLength, iData.begin());
	}

	packet::~packet()
	{
	}

	void packet::set(int aId, char aData)
	{
		set(aId, sizeof(char), &aData);
	}

	void packet::set(int aId, int aData)
	{
		set(aId, sizeof(int), &aData);
	}

	void packet::set(int aId, size_t aLength, const void* aData)
	{
		iId = aId;
		iData.resize(aLength);
		if (aData)
			std::copy(static_cast<const char*>(aData), static_cast<const char*>(aData) + aLength, iData.begin());
	}

	void packet::set(size_t aLength, const void* aData)
	{
		set(0, aLength, aData);
	}

	void packet::swap(packet& aOther)
	{
		std::swap(iId, aOther.iId);
		iData.swap(aOther.iData);
	}

	template <>
	std::string packet_value<std::string>(const packet& aPacket)
	{
		std::string ret;
		ret.assign(reinterpret_cast<const char*>(aPacket.data()), aPacket.length());
		return ret;
	}

	void save_to_packet(packet& aPacket, serializable& aObject)
	{
		size_t length;
		aObject.save(lib::len_outstream(length));
		aPacket.set(aPacket.id(), length, 0);
		aObject.save(lib::ptr_outstream(aPacket.data(), length));
	}

	void load_from_packet(const packet& aPacket, serializable& aObject)
	{
		aObject.load(lib::ptr_instream(aPacket.data(), aPacket.length()));
	}

	struct socket_error
	{
		int iErrorCode;
		const char* iErrorString;
	} sSocketErrors[] =
	{
		{ WSAEACCES, "Permission denied." },
		{ WSAEADDRINUSE, "Address already in use." },
		{ WSAEADDRNOTAVAIL, "Cannot assign requested address." },
		{ WSAEAFNOSUPPORT, "Address family not supported by protocol family." },
		{ WSAEALREADY, "Operation already in progress." },
		{ WSAECONNABORTED, "Software caused connection abort." },
		{ WSAECONNREFUSED, "Connection refused." },
		{ WSAECONNRESET, "Connection reset by peer." },
		{ WSAEDESTADDRREQ, "Destination address required." },
		{ WSAEFAULT, "Bad address." },
		{ WSAEHOSTDOWN, "Host is down." },
		{ WSAEHOSTUNREACH, "No route to host." },
		{ WSAEINPROGRESS, "Operation now in progress." },
		{ WSAEINTR, "Interrupted function call." },
		{ WSAEINVAL, "Invalid argument." },
		{ WSAEISCONN, "Socket is already connected." },
		{ WSAEMFILE, "Too many open files." },
		{ WSAEMSGSIZE, "Message too long." },
		{ WSAENETDOWN, "Network is down." },
		{ WSAENETRESET, "Network dropped connection on reset." },
		{ WSAENETUNREACH, "Network is unreachable." },
		{ WSAENOBUFS, "No buffer space available." },
		{ WSAENOPROTOOPT, "Bad protocol option." },
		{ WSAENOTCONN, "Socket is not connected." },
		{ WSAENOTSOCK, "Socket operation on non-socket." },
		{ WSAEOPNOTSUPP, "Operation not supported." },
		{ WSAEPFNOSUPPORT, "Protocol family not supported." },
		{ WSAEPROCLIM, "Too many processes." },
		{ WSAEPROTONOSUPPORT, "Protocol not supported." },
		{ WSAEPROTOTYPE, "Protocol wrong type for socket." },
		{ WSAESHUTDOWN, "Cannot send after socket shutdown." },
		{ WSAESOCKTNOSUPPORT, "Socket type not supported." },
		{ WSAETIMEDOUT, "Connection timed out." },
		{ WSATYPE_NOT_FOUND, "Class type not found." },
		{ WSAEWOULDBLOCK, "A non-blocking socket operation could not be completed immediately." },
		{ WSAHOST_NOT_FOUND, "Host not found." },
		{ WSA_INVALID_HANDLE, "Specified event object handle is invalid." },
		{ WSA_INVALID_PARAMETER, "One or more parameters are invalid." },
		{ WSAEINVALIDPROCTABLE, "Invalid procedure table from service provider." },
		{ WSAEINVALIDPROVIDER, "Invalid service provider version number." },
		{ WSA_IO_INCOMPLETE, "Overlapped I/O event object not in signaled state." },
		{ WSA_IO_PENDING, "Overlapped operations will complete later." },
		{ WSA_NOT_ENOUGH_MEMORY, "Insufficient memory available." },
		{ WSANOTINITIALISED, "Successful WSAStartup not yet performed." },
		{ WSANO_DATA, "Valid name, no data record of requested type." },
		{ WSANO_RECOVERY, "This is a non-recoverable error." },
		{ WSAEPROVIDERFAILEDINIT, "Unable to initialize a service provider." },
		{ WSASYSCALLFAILURE, "System call failure." },
		{ WSASYSNOTREADY, "Network subsystem is unavailable." },
		{ WSATRY_AGAIN, "Non-authoritative host not found." },
		{ WSAVERNOTSUPPORTED, "WINSOCK.DLL version out of range." }
	};

	socket::socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol) : 
		iOwner(aOwner),
		iSocketIsBlocking(true), iBlocking(aBlocking), iProtocol(aProtocol), iHandle(INVALID_SOCKET), iParent(0), iError(0),
		iAddress(INADDR_NONE), iPort(0), iConnected(false),
		iInputBufferRecvPtr(iInputBuffer), iInputBufferNextPacketPtr(iInputBuffer), 
		iOutputBufferNextPacketPtr(iOutputBuffer),
		iStatus(StatusNone), iUnderflow(false), iHasUnderflow(false)
	{
	}

	socket::socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, SOCKET aHandle, bool aConnected) : 
		iOwner(aOwner),
		iSocketIsBlocking(true), iBlocking(aBlocking), iProtocol(aProtocol), iHandle(aHandle), iParent(0), iError(0),
		iAddress(INADDR_NONE), iPort(0), iConnected(aConnected),
		iInputBufferRecvPtr(iInputBuffer), iInputBufferNextPacketPtr(iInputBuffer), 
		iOutputBufferNextPacketPtr(iOutputBuffer),
		iStatus(StatusNone), iUnderflow(false), iHasUnderflow(false)
	{
	}

	socket::socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, SOCKET aHandle, u_long aAddress, u_short aPort, bool aConnected) : 
		iOwner(aOwner),
		iSocketIsBlocking(true), iBlocking(aBlocking), iProtocol(aProtocol), iHandle(aHandle), iParent(0), iError(0),
		iAddress(aAddress), iPort(aPort), iConnected(aConnected),
		iInputBufferRecvPtr(iInputBuffer), iInputBufferNextPacketPtr(iInputBuffer), 
		iOutputBufferNextPacketPtr(iOutputBuffer),
		iStatus(StatusNone), iUnderflow(false), iHasUnderflow(false)
	{
	}

	socket::socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, u_long aAddress, u_short aPort, bool aConnect) : 
		iOwner(aOwner),
		iSocketIsBlocking(true), iBlocking(aBlocking), iProtocol(aProtocol), iHandle(INVALID_SOCKET), iParent(0), iError(0),
		iAddress(aAddress), iPort(aPort), iConnected(false),
		iInputBufferRecvPtr(iInputBuffer), iInputBufferNextPacketPtr(iInputBuffer), 
		iOutputBufferNextPacketPtr(iOutputBuffer),
		iStatus(StatusNone), iUnderflow(false), iHasUnderflow(false)
	{
		if (aConnect)
			connect();
	}

	socket::~socket()
	{
		if (iHandle != INVALID_SOCKET)
			close();
	}

	bool socket::create(bool aCoalesce)
	{
		if (iHandle != INVALID_SOCKET)
			close();

		if ((iHandle = ::socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
		{
			get_error();
			return false;
		}
		
		if (!aCoalesce)
		{
			int optval = 1;
			if (setsockopt(iHandle, IPPROTO_TCP, TCP_NODELAY, 
				reinterpret_cast<char*>(&optval), sizeof(optval)) == SOCKET_ERROR)
			{
				get_error();
				close();
				return false;
			}
		}

		if (!set_blocking())
		{
			get_error();
			close();
			return false;
		}

		return true;
	}

	bool socket::connect(bool aCoalesce)
	{
		if (iConnected)
			close();

		if (iHandle == INVALID_SOCKET && !create(aCoalesce))
			return false;

		sockaddr_in sa;
		sa.sin_family = AF_INET;
		sa.sin_addr.s_addr = htonl(iAddress);
		sa.sin_port = htons(iPort);

		if (::connect(iHandle, reinterpret_cast<sockaddr*>(&sa), sizeof(sa)))
		{
			if (::WSAGetLastError() != WSAEWOULDBLOCK)
			{
				get_error();
				closesocket(iHandle);
				iHandle = INVALID_SOCKET;
				return false;
			}
		}
		else
			set_connected();

		return true;
	}

	void socket::close()
	{
		if (iHandle == INVALID_SOCKET)
			return;

		iOwner.lock();
		closesocket(iHandle);
		iHandle = INVALID_SOCKET;
		set_connected(false);
		iOwner.unlock();
	}

	u_long socket::local_address() const
	{
		sockaddr_in sa; 
		int size = sizeof(sa);
		if (getsockname(iHandle, reinterpret_cast<sockaddr*>(&sa), &size) == INVALID_SOCKET)
			return 0;
		return ntohl(sa.sin_addr.s_addr);
	}

	u_short socket::local_port() const
	{
		sockaddr_in sa; 
		int size = sizeof(sa);
		if (getsockname(iHandle, reinterpret_cast<sockaddr*>(&sa), &size) == INVALID_SOCKET)
			return 0;
		return ntohs(sa.sin_port);
	}

	bool socket::set_blocking(blocking_type aBlocking)
	{
		iBlocking = aBlocking;

		if (iBlocking == Blocking && iSocketIsBlocking)
			return true;

		if (iBlocking == NonBlocking && !iSocketIsBlocking)
			return true;

		u_long arg = (iBlocking == Blocking ? 0 : 1);
		if (ioctlsocket(iHandle, FIONBIO, &arg) == 0)
		{
			iSocketIsBlocking = (arg == 0 ? true : false);
			return true;
		}
		else
		{
			iBlocking = (iSocketIsBlocking ? Blocking : NonBlocking);
			get_error();
			return false;
		}
	}

	void socket::set_protocol(protocol_type aProtocol) 
	{ 
		if (iProtocol == aProtocol)
			return;
		
		iProtocol = aProtocol;
	}

	void socket::output(const packet& aOutputPacket)
	{
		iOwner.lock();
		iOutputPackets.push_back(aOutputPacket);
		iOwner.unlock();
	}

	bool socket::has_output() const
	{ 
		return !iOutputPackets.empty() || iOutputBufferNextPacketPtr > iOutputBuffer; 
	}

	void socket::get_error()
	{
		set_error(WSAGetLastError());
		if (iHandle != INVALID_SOCKET)
		{
			int error = 0;
			int optlen = sizeof(error);
			if (getsockopt(iHandle, SOL_SOCKET, SO_ERROR, reinterpret_cast<char*>(&error), &optlen) == 0)
			{
				if (error != 0)
					set_error(error);
			}
		}
	}

	void socket::set_error(int aError)
	{
		iError = aError;
		iErrorString = "";
		for (size_t i = 0; i < sizeof(sSocketErrors) / sizeof(sSocketErrors[0]); ++i)
		{
			if (sSocketErrors[i].iErrorCode == iError)
			{
				iErrorString = sSocketErrors[i].iErrorString;
				break;
			}
		}
	}
	
	bool socket::read(char* aBuffer, int aLength, int& aBytesRead)
	{
		aBytesRead = recv(iHandle, aBuffer, aLength, 0);
		if (aBytesRead == 0)
		{
			close();
			return true;
		}
		else if (aBytesRead < 0) 
		{
			get_error();
			close();
			return false;
		}
		return true;
	}

	bool socket::write(const char* aBuffer, int aLength, int& aBytesWritten)
	{
		aBytesWritten = send(iHandle, aBuffer, aLength, 0);
		if (aBytesWritten == SOCKET_ERROR)
		{
			aBytesWritten = 0;
			get_error();
			close();
			return false;
		}
		return true;
	}

	void socket::set_connected(bool aConnected)
	{ 
		if (iConnected == aConnected)
			return;
		iConnected = aConnected;
		if (iConnected)
			handle_connected();
		else
			handle_disconnected();
	}
	 
	bool socket::handle_input()
	{
		if (try_write_again(Read))
			return handle_output();
		do
		{
			int br;
			bool ok = read(iInputBufferRecvPtr, (iInputBuffer + SocketBufferLength) - iInputBufferRecvPtr, br);
			if (!ok || !connected())
				return ok;
			if (br == 0)
				continue;
			if (iProtocol == Packet)
			{
				size_t bytesLeft = (iInputBufferRecvPtr + br) - iInputBufferNextPacketPtr;
				while(bytesLeft >= PacketHeaderLength)
				{
					int id = static_cast<int>(ntohl(static_cast<u_long>(*((int*) iInputBufferNextPacketPtr))));
					size_t length = static_cast<size_t>(ntohl(static_cast<u_long>(*reinterpret_cast<size_t*>(iInputBufferNextPacketPtr + sizeof(int)))));
					if (bytesLeft - PacketHeaderLength >= length)
					{
						if (id == 0 && length == 0) // combine split packet
						{
							bool first = true;
							packet_list::iterator largePacket;
							char* largePacketData;
							for (packet_list::iterator i = iInputPackets.begin(); i != iInputPackets.end();)
							{
								if (i->id() == 0)
								{
									if (first)
									{
										first = false;
										largePacket = iInputPackets.insert(i, packet(0, 0, 0));
										largePacket->set(static_cast<int>(ntohl(static_cast<u_long>(*reinterpret_cast<int*>(i->data())))), static_cast<size_t>(ntohl(static_cast<u_long>(*reinterpret_cast<size_t*>(static_cast<char*>(i->data()) + sizeof(int))))), 0);
										largePacketData = static_cast<char*>(largePacket->data());
										memcpy(largePacketData, static_cast<const char*>(i->data()) + PacketHeaderLength, i->length() - PacketHeaderLength);
										largePacketData += i->length() - PacketHeaderLength;
									}
									else
									{
										memcpy(largePacketData, i->data(), i->length());
										largePacketData += i->length();
									}
									iInputPackets.erase(i++);
								}
								else
									++i;
							}
						}
						else
						{
							iInputPackets.push_back(packet());
							iInputPackets.back().set(id, length, iInputBufferNextPacketPtr + PacketHeaderLength);
						}
						iInputBufferNextPacketPtr += length + PacketHeaderLength;
						bytesLeft = (iInputBufferRecvPtr + br) - iInputBufferNextPacketPtr;
					}
					else
						break;
				}
				memmove(iInputBuffer, iInputBufferNextPacketPtr, bytesLeft);
				iInputBufferNextPacketPtr = iInputBuffer;
				iInputBufferRecvPtr = iInputBuffer + bytesLeft;

				for (packet_list::iterator i = iInputPackets.begin(); i != iInputPackets.end();)
				{
					packet& thePacket = *i;
					if (thePacket.id() != 0)
					{
						iOwner.add_notification(socket_manager_observer::NotifyData, *this, thePacket);
						iInputPackets.erase(i++);
					}
					else
						++i;
				}
			}
			else
			{
				iBasicProtocolPacket.set(0, br, iInputBufferRecvPtr);
				if (iBasicProtocolPacket.length() > 0)
					iOwner.add_notification(socket_manager_observer::NotifyData, *this, iBasicProtocolPacket);
			}
		} while (read_more());
		return true;
	}

	bool socket::handle_output()
	{	
		if (try_read_again(Write))
			return handle_input();
		size_t bytesLeft = (iOutputBuffer + SocketBufferLength) - iOutputBufferNextPacketPtr;
		if (!try_read_again(Write) && !try_write_again(Read))
		{
			iUnderflow = true;
			size_t headerSize = (iProtocol == Packet ? PacketHeaderLength : 0); 
			while(bytesLeft >= headerSize && !iOutputPackets.empty())
			{
				iUnderflow = false;
				if (iOutputPackets.front().length() > MaxPacketLength) // split large packets up
				{
					packet_list::iterator largePacket = iOutputPackets.begin();
					char headerPacket[MaxPacketLength];
					*reinterpret_cast<int*>(headerPacket) = static_cast<int>(htonl(static_cast<u_long>(largePacket->id())));
					*reinterpret_cast<size_t*>(headerPacket + sizeof(int)) = static_cast<size_t>(htonl(static_cast<u_long>(largePacket->length())));
					memcpy(headerPacket + headerSize, largePacket->data(), MaxPacketLength - headerSize);
					iOutputPackets.insert(largePacket, packet(0, MaxPacketLength, headerPacket));
					size_t largePacketBytesLeft = largePacket->length() - (MaxPacketLength - headerSize);
					const char* largePacketData = static_cast<const char*>(largePacket->data()) + (MaxPacketLength - headerSize);
					while (largePacketBytesLeft > 0)
					{
						size_t nextSize = largePacketBytesLeft > MaxPacketLength ? MaxPacketLength : largePacketBytesLeft;
						iOutputPackets.insert(largePacket, packet(0, nextSize, largePacketData));
						largePacketData += nextSize;
						largePacketBytesLeft -= nextSize;
					}
					iOutputPackets.insert(largePacket, packet(0, 0, 0));
					iOutputPackets.erase(largePacket);
				}

				if (iProtocol == Packet)
				{
					int id = iOutputPackets.front().id();
					size_t length = iOutputPackets.front().length();
					if (bytesLeft - PacketHeaderLength >= length)
					{
						*((int*) iOutputBufferNextPacketPtr) =  static_cast<int>(htonl(static_cast<u_long>(id)));
						*((size_t*) iOutputBufferNextPacketPtr + 1) = static_cast<size_t>(htonl(static_cast<u_long>(length)));
						if (length > 0)
							memcpy(iOutputBufferNextPacketPtr + PacketHeaderLength, iOutputPackets.front().data(), length);
						iOutputPackets.pop_front();
						iOutputBufferNextPacketPtr += length + PacketHeaderLength;
						bytesLeft = (iOutputBuffer + SocketBufferLength) - iOutputBufferNextPacketPtr;
					}
					else
						break;
				}
				else // iProtocol == Basic
				{
					size_t length = iOutputPackets.front().length();
					if (bytesLeft >= length)
					{
						if (length > 0)
							memcpy(iOutputBufferNextPacketPtr, iOutputPackets.front().data(), length);
						iOutputPackets.pop_front();
						iOutputBufferNextPacketPtr += length;
						bytesLeft = (iOutputBuffer + SocketBufferLength) - iOutputBufferNextPacketPtr;
					}
					else
						break;
				}
			}
		}
		
		if (iOutputBufferNextPacketPtr - iOutputBuffer == 0)
			return true;

		int bs;
		bool ok = write(iOutputBuffer, iOutputBufferNextPacketPtr - iOutputBuffer, bs);
		if (!ok || !connected() || bs == 0)
			return ok;
		else
		{
			bytesLeft = (iOutputBufferNextPacketPtr - iOutputBuffer) - bs;
			if (bytesLeft > 0)
				memmove(iOutputBuffer, iOutputBuffer + bs, bytesLeft);
			iOutputBufferNextPacketPtr -= bs;
	
			if (iUnderflow && iHasUnderflow)
				iOwner.add_notification(socket_manager_observer::NotifyUnderflow, *this);
		}

		return ok;
	}

	bool socket::handle_error()
	{
		get_error();
		close();
		return false;
	}

	
	listen_socket::listen_socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, u_short aPort) : 
		socket(aOwner, aBlocking, aProtocol, 0, aPort, false)
	{
		listen();
	}

	bool listen_socket::handle_input()
	{
		SOCKET s;
		sockaddr_in sa; int size = sizeof(sa);
		if ((s = accept(iHandle, reinterpret_cast<sockaddr*>(&sa), &size)) == INVALID_SOCKET)
		{
			get_error();
			close();
			return false;
		}
		if (getpeername(s, reinterpret_cast<sockaddr*>(&sa), &size) == INVALID_SOCKET)
		{
			get_error();
			close();
			closesocket(s);
			return false;
		}
		int opt = 1;
		setsockopt(s, SOL_SOCKET, SO_DONTLINGER, reinterpret_cast<char*>(&opt), sizeof(opt));

		socket* newSocket = new socket(iOwner, iBlocking, iProtocol, s, ntohl(sa.sin_addr.s_addr), ntohs(sa.sin_port), true);
		newSocket->set_parent(this);
		iOwner.add_socket(newSocket);

		return true;
	}

	bool listen_socket::listen()
	{
		if (iHandle != INVALID_SOCKET)
			close();
		
		if ((iHandle = ::socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
			return false;

		int optval = 1;
		if (setsockopt(iHandle, SOL_SOCKET, SO_REUSEADDR, 
			reinterpret_cast<char*>(&optval), sizeof(optval)) == SOCKET_ERROR)
		{
			get_error();
			close();
			return false;
		}

		sockaddr_in sa = {};
		sa.sin_family = AF_INET;
		sa.sin_port = htons(iPort);

		if (bind(iHandle, reinterpret_cast<sockaddr*>(&sa), sizeof(sa)) == SOCKET_ERROR)
		{
			get_error();
			close();
			return false;
		}

		if (!set_blocking())
		{
			get_error();
			close();
			return false;
		}

		if (::listen(iHandle, 256) == SOCKET_ERROR)
		{
			get_error();
			close();
			return false;
		}

		set_connected();

		return true;
	}

	#ifndef NO_SSL

	ssl_socket::ssl_socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol) : socket(aOwner, aBlocking, aProtocol), iSsl(0), iBio(0), iReadSsl(None), iWriteSsl(None)
	{
	}

	ssl_socket::ssl_socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, SOCKET aHandle, bool aConnected) : socket(aOwner, aBlocking, aProtocol, aHandle, aConnected), iSsl(0), iBio(0), iReadSsl(None), iWriteSsl(None)
	{
	}

	ssl_socket::ssl_socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, SOCKET aHandle, u_long aAddress, u_short aPort, bool aConnected) : socket(aOwner, aBlocking, aProtocol, aHandle, aAddress, aPort, aConnected), iSsl(0), iBio(0), iReadSsl(None), iWriteSsl(None)
	{
	}

	ssl_socket::ssl_socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, u_long aAddress, u_short aPort, bool aConnect) : socket(aOwner, aBlocking, aProtocol, aAddress, aPort, aConnect), iSsl(0), iBio(0), iReadSsl(None), iWriteSsl(None)
	{
	}

	ssl_socket::~ssl_socket()
	{
		if (iSsl != 0)
		{
			if (SSL_shutdown(iSsl) == 0)
				SSL_shutdown(iSsl);
			SSL_free(iSsl);
			iSsl = 0;
			iBio = 0;
			iReadSsl = iWriteSsl = None;
		}
	}

	bool ssl_socket::initializing() const
	{
		return SSL_in_init(iSsl) && !SSL_total_renegotiations(iSsl);
	}

	bool ssl_socket::pending() const
	{
		return SSL_pending(iSsl) != 0;
	}

	bool ssl_socket::has_output() const
	{
		return (socket::has_output() || iReadSsl == Write || iWriteSsl == Write)
			&& iWriteSsl != Read;
	}

	void ssl_socket::handle_connected()
	{
		assert(iSsl == 0);
		iSsl = SSL_new(iOwner.ssl_ctx());
		iBio = BIO_new_socket(iHandle, BIO_NOCLOSE);
		BIO_set_nbio(iBio, iBlocking == Blocking ? 0 : 1);
		SSL_set_bio(iSsl,iBio,iBio);
		SSL_set_connect_state(iSsl);
	}

	void ssl_socket::handle_disconnected()
	{
		if (iSsl != 0)
		{
			if (SSL_shutdown(iSsl) == 0)
				SSL_shutdown(iSsl);
			SSL_free(iSsl);
			iSsl = 0;
			iBio = 0;
			iReadSsl = iWriteSsl = None;
		}
	}

	bool ssl_socket::read(char* aBuffer, int aLength, int& aBytesRead)
	{
		aBytesRead = 0;
		iReadSsl = None;
		int r = SSL_read(iSsl, aBuffer, aLength);
		int err = SSL_get_error(iSsl, r);
		switch(err)
		{
		case SSL_ERROR_NONE:
			aBytesRead = r;
			break;
		case SSL_ERROR_ZERO_RETURN:
			close();
			return true;
		case SSL_ERROR_WANT_READ:
			iReadSsl = Read;
			break;
		case SSL_ERROR_WANT_WRITE:
			iReadSsl = Write;
			break;
		default:
			get_error();
			close();
			return false;
		}
		return true;
	}

	bool ssl_socket::read_more() const
	{
		return pending();
	}

	bool ssl_socket::write(const char* aBuffer, int aLength, int& aBytesWritten)
	{
		aBytesWritten = 0;
		iWriteSsl = None;
		int r = SSL_write(iSsl, aBuffer, aLength);
		int err = SSL_get_error(iSsl, r);
		switch(err)
		{
		case SSL_ERROR_NONE:
			aBytesWritten = r;
			break;
		case SSL_ERROR_ZERO_RETURN:
			close();
			return true;
		case SSL_ERROR_WANT_WRITE:
			iWriteSsl = Write;
			break;
		case SSL_ERROR_WANT_READ:
			iWriteSsl = Read;
			break;
		default:	      
			get_error();
			close();
			return false;
		}
		return true;
	}
	
	#endif //NO_SSL

	resolver::resolver()
	{
	}

	resolver::~resolver()
	{
		for (entries::iterator i = iEntries.begin(); i != iEntries.end(); ++i)
			i->abort();
	}

	void resolver::add(const std::string& aHostName, bool aDoReverseLookup)
	{
		iEntries.push_back(entry(*this));
		iEntries.back().iHostName = aHostName;
		iEntries.back().iDoReverseLookup = aDoReverseLookup;
		iEntries.back().run();
	}

	void resolver::add(const socket& aSocket, const std::string& aHostName, bool aDoReverseLookup)
	{
		iEntries.push_back(entry(*this, aSocket));
		iEntries.back().iHostName = aHostName;
		iEntries.back().iDoReverseLookup = aDoReverseLookup;
		iEntries.back().run();
	}

	void resolver::add(requester& aRequester, const std::string& aHostName, bool aDoReverseLookup)
	{
		iEntries.push_back(entry(*this, aRequester));
		iEntries.back().iHostName = aHostName;
		iEntries.back().iDoReverseLookup = aDoReverseLookup;
		iEntries.back().run();
	}

	void resolver::remove(const socket& aSocket)
	{
		for (entries::iterator i = iEntries.begin(); i != iEntries.end();)
		{
			if (i->iSocket == &aSocket)
			{
				i->abort();
				iEntries.erase(i++);
			}
			else
				++i;
		}
	}

	void resolver::remove(requester& aRequester)
	{
		for (entries::iterator i = iEntries.begin(); i != iEntries.end();)
		{
			if (i->iRequester == &aRequester)
			{
				i->abort();
				iEntries.erase(i++);
			}
			else
				++i;
		}
	}

	bool resolver::is_next_ready() const
	{
		for (entries::const_iterator i = iEntries.begin(); i != iEntries.end(); ++i)
		{
			if (i->is_finished())
				return true;
		}
		return false;
	}

	bool resolver::next_ready(entry& aEntry)
	{
		for (entries::iterator i = iEntries.begin(); i != iEntries.end(); ++i)
		{
			if (i->is_finished())
			{
				aEntry = *i;
				iEntries.erase(i);
				return true;
			}
		}
		return false;
	}

	resolver::entry& resolver::entry::operator=(const entry& aOther)
	{
		iSocket = aOther.iSocket;
		iRequester = aOther.iRequester;
		iHostName = aOther.iHostName;
		iHostNumber = aOther.iHostNumber;
		iDoReverseLookup = aOther.iDoReverseLookup;
		iAddress = aOther.iAddress;
		iReady = aOther.iReady;
		iError = aOther.iError;
		return *this;
	}

	void resolver::entry::task()
	{
		int error = 0;
		u_long address = INADDR_NONE;
		if (!iHostName.empty())
		{
			lock();
			address = inet_addr(iHostName.c_str());
			unlock();
			if (address != INADDR_NONE)
			{
				address = ntohl(address);
				if (iDoReverseLookup)
				{
				    struct sockaddr_in saGNI;
				    char hostname[NI_MAXHOST];
					char servInfo[NI_MAXSERV];
					hostname[0] = 0;
					u_short port = 27015;
					saGNI.sin_family = AF_INET;
					saGNI.sin_addr.s_addr = htonl(address);
					saGNI.sin_port = htons(port);
					if (getnameinfo((struct sockaddr *) &saGNI,
							   sizeof (struct sockaddr),
							   hostname,
							   NI_MAXHOST, servInfo, 
							   NI_MAXSERV, NI_NUMERICSERV) != 0)
						error = ::WSAGetLastError();
					lock();
					iHostName = hostname;
					unlock();
				}
			}
			else
			{
				hostent *he = gethostbyname(iHostName.c_str());
				if (he == NULL)
					error = ::WSAGetLastError();
				else
				{
					sockaddr_in sa;
					memcpy(&sa.sin_addr, he->h_addr, he->h_length);
					address = ntohl(sa.sin_addr.s_addr);
				}
			}
		}
		iAddress = address;
		if (address != INADDR_NONE)
		{
			in_addr ia;
			ia.s_addr = htonl(address);
			lock();
			iHostNumber = inet_ntoa(ia);
			unlock();
		}
		else
		{
			lock();
			iHostNumber = "";
			unlock();
		}
		iError = error;
		iReady = true;
		return;
	}

	namespace
	{
		LRESULT __stdcall processorWndProc(HWND aWindow, UINT aMessage, WPARAM aWParam, LPARAM aLParam)
		{
			switch (aMessage)
			{
			case WM_NULL:
			case WM_TIMER:
				reinterpret_cast<processor*>(GetWindowLongPtrA(aWindow, GWLP_USERDATA))->process();
				return 0;
			default:
				return ::DefWindowProcA(aWindow, aMessage, aWParam, aLParam);
			}
		};
	}

	#include <windows.h>
	processor::processor(socket_manager& aOwner) : iOwner(aOwner), iHasWork(false)
	{
		if (iOwner.processor_type() == socket_manager::Internal)
		{
			WNDCLASSEXA wcxNarrow;
			wcxNarrow.cbSize = sizeof(wcxNarrow);
			wcxNarrow.style = CS_PARENTDC|CS_GLOBALCLASS|CS_DBLCLKS;
			wcxNarrow.lpfnWndProc = processorWndProc;
			wcxNarrow.cbClsExtra = 0;
			wcxNarrow.cbWndExtra = 0;
			wcxNarrow.hInstance = (HINSTANCE) iOwner.application_handle();
			wcxNarrow.hIcon = NULL;
			wcxNarrow.hCursor = NULL;
			wcxNarrow.hbrBackground = NULL;
			wcxNarrow.lpszMenuName = NULL;
			wcxNarrow.lpszClassName = "i42SoftwareNetworkHelper";
			wcxNarrow.hIconSm = NULL;
			::RegisterClassExA(&wcxNarrow);
			iSink = ::CreateWindowA("i42SoftwareNetworkHelper", "", WS_POPUP, 0, 0, 0, 0, NULL, NULL, (HINSTANCE) iOwner.application_handle(), 0);
			::SetWindowLongPtrA(iSink, GWLP_USERDATA, (LONG_PTR)this);
			::SetTimer(iSink, 1, 10, NULL);
		}
	}

	processor::~processor()
	{
		if (iOwner.processor_type() == socket_manager::Internal)
		{
			::KillTimer(iSink, 1);
			::DestroyWindow(iSink);
			::UnregisterClassA("i42SoftwareNetworkHelper", (HINSTANCE) iOwner.application_handle());
		}
	}

	void processor::task()
	{
		iOwner.lock();
		if (!iHasWork)
		{
			if (iOwner.processor_type() == socket_manager::Internal)
				::PostMessage(iSink, WM_NULL, 0, 0);
			else
				iEvent.signal();
			iHasWork = true;
		}
		iOwner.unlock();
	}
	
	void processor::process()
	{
		iOwner.lock();
		iOwner.process();
		iHasWork = false;
		iOwner.unlock();
	}

	bool processor::wait(DWORD aTimeOut)
	{
		return iEvent.msg_wait(aTimeOut) != 0;
	}

	socket_manager::socket_manager(void* aApplicationHandle, processor_type_e aProcessorType, socket::blocking_type aBlocking, socket::protocol_type aProtocol) : 
		iApplicationHandle(aApplicationHandle),
		iProcessorType(aProcessorType), 
		iProcessor(*this), iYieldProc(0), iProcessing(false), iBlocking(aBlocking), iProtocol(aProtocol), iError(0), iInitialised(false)
			#ifndef NO_SSL
			, iSslStarted(false)
			#endif //NO_SSL
	{
		WSADATA wsaData;
		if (::WSAStartup(MAKEWORD(2, 0), &wsaData) == 0)
		{
			iInitialised = true;
			run();
		}
		else
			get_error();
	}

	socket_manager::~socket_manager()
	{
		abort();

		for (socket_list::iterator i = iSockets.begin(); i != iSockets.end(); ++i)
		{
			delete *i;
		}
		iSockets.erase(iSockets.begin(), iSockets.end());
	
		#ifndef NO_SSL
		close_ssl();
		#endif

		if (iInitialised)
			::WSACleanup();
	}

	#ifndef NO_SSL
	void socket_manager::close_ssl()
	{
		if (iSslStarted)
		{
			iSslStarted = 0;
			SSL_CTX_free(iSslCtx);
			ERR_remove_state(0);
			ENGINE_cleanup();
			CONF_modules_unload(1);
			CRYPTO_cleanup_all_ex_data();
			ERR_free_strings();
			EVP_cleanup();
		}
	}
	#endif //NO_SSL

	socket* socket_manager::listen(u_short aPort)
	{
		listen_socket* theSocket = new listen_socket(*this, iBlocking, iProtocol, aPort);
		if (!theSocket->listen())
		{
			get_error();
			delete theSocket;
			return 0;
		}
	
		add_socket(theSocket, true);
		return theSocket;
	}

	socket* socket_manager::connect(u_long aAddress, u_short aPort, bool aSecure)
	{
		socket* theSocket = 0;
		#ifndef NO_SSL
		if (!aSecure)
			theSocket = new socket(*this, iBlocking, iProtocol, aAddress, aPort, false);
		else
			theSocket = new ssl_socket(*this, iBlocking, iProtocol, aAddress, aPort, false);
		#else
		theSocket = new socket(*this, iBlocking, iProtocol, aAddress, aPort, false);
		#endif //NO_SSL
		if (!theSocket->connect())
		{
			iError = theSocket->error();
			iErrorString = theSocket->error_string();
			delete theSocket;
			theSocket = 0;
		}
		else
		{
			add_socket(theSocket);
			if (theSocket->connected())
				notify_observers(socket_manager_observer::NotifySocketConnected, *theSocket);
		}
		return theSocket;
	}

	socket* socket_manager::connect(const std::string& aHostName, u_short aPort, bool aSecure)
	{
		socket* theSocket = 0;

		u_long address = to_address(aHostName);

		if (address != INADDR_NONE)
			theSocket = connect(address, aPort, aSecure);
		else if (iBlocking == socket::NonBlocking)
		{
			#ifndef NO_SSL
			if (!aSecure)
				theSocket = new socket(*this, iBlocking, iProtocol, static_cast<u_long>(INADDR_NONE), aPort, false);
			else
				theSocket = new ssl_socket(*this, iBlocking, iProtocol, static_cast<u_long>(INADDR_NONE), aPort, false);
			#else //NO_SSL
			theSocket = new socket(*this, iBlocking, iProtocol, static_cast<u_long>(INADDR_NONE), aPort, false);
			#endif
			if (!theSocket->create())
			{
				get_error();
				delete theSocket;
				theSocket = 0;
			}
			else
			{
				add_socket(theSocket);
				iResolver.add(*theSocket, aHostName);
			}
		}
		else
			throw std::logic_error(std::string("lib::socket_manager::connect"));

		return theSocket;
	}

	void socket_manager::task()
	{
		struct on_exit
		{
			socket_manager& iParent;
			on_exit(socket_manager& aParent) : iParent(aParent) {}
			~on_exit()
			{
				#ifndef NO_SSL
				iParent.close_ssl();
				#endif //NO_SSL
			}
		} onExit(*this);
		while(is_running())
		{
			lock();

			while (iSockets.empty() && is_running())
			{
				unlock();
				Sleep(100);
				if (!is_running())
					return;
				lock();
			}

			iReadFds.clear();
			iWriteFds.clear();
			iExceptFds.clear();

			for (socket_list::iterator i = iSockets.begin(); i != iSockets.end(); ++i)
			{
				socket& theSocket = **i;

				if (theSocket.address() == INADDR_NONE)
					continue;

				if (theSocket.zombie())
					continue;

				SOCKET handle = theSocket.handle();
				iReadFds[&theSocket] = std::make_pair(false, handle);
				if (theSocket.has_output() || theSocket.has_underflow() || !theSocket.connected())
					iWriteFds[&theSocket] = std::make_pair(false, handle);
				iExceptFds[&theSocket] = std::make_pair(false, handle);
			}

			bool finished = iReadFds.empty() && iWriteFds.empty() && iExceptFds.empty();
			while(!finished)
			{
				finished = true;

				fd_set readfds;
				fd_set writefds;
				fd_set exceptfds;

				FD_ZERO(&readfds);
				FD_ZERO(&writefds);
				FD_ZERO(&exceptfds);

				std::size_t count = 0;
				for (fd_list::iterator i = iReadFds.begin(); i != iReadFds.end(); ++i)
				{
					if (!i->second.first)
					{
						finished = false;
						i->second.first = true;
						FD_SET(i->second.second, &readfds);
						if (++count == FD_SETSIZE)
							break;
					}
				}
				count = 0;
				for (fd_list::iterator i = iWriteFds.begin(); i != iWriteFds.end(); ++i)
				{
					if (!i->second.first)
					{					
						finished = false;
						i->second.first = true;
						FD_SET(i->second.second, &writefds);
						if (++count == FD_SETSIZE)
							break;
					}
				}
				count = 0;
				for (fd_list::iterator i = iExceptFds.begin(); i != iExceptFds.end(); ++i)
				{
					if (!i->second.first)
					{					
						finished = false;
						i->second.first = true;
						FD_SET(i->second.second, &exceptfds);
						if (++count == FD_SETSIZE)
							break;
					}
				}

				if (!finished)
				{
					if (!do_select(readfds, writefds, exceptfds))
					{
						unlock();
						return;
					}
					do_io();
				}
			}
			unlock();
		}
	}

	void socket_manager::process_any(DWORD aTimeOut)
	{
		assert(iProcessorType == External);
		do
		{
			process();
		}
		while (!yield() && iProcessor.wait(aTimeOut));
	}

	void socket_manager::process()
	{
		if (iProcessing)
			return;
		lock();
		iProcessing = true;
		process_notifications();
		process_zombies();
		process_resolved();
		iProcessing = false;
		unlock();
	}

	bool socket_manager::connected() const
	{
		for (socket_list::const_iterator i = iSockets.begin(); i != iSockets.end(); ++i)
			if ((*i)->connected())
				return true;
		return false;
	}

	void socket_manager::disconnect()
	{
		for (socket_list::iterator i = iSockets.begin(); i != iSockets.end(); ++i)
			(*i)->close();
	}

	void socket_manager::disconnect(socket& aSocket)
	{
		aSocket.close();
	}

	#ifndef NO_SSL
	void socket_manager::start_ssl()
	{
		if (iSslStarted)
			return;
		iSslStarted = true;
		SSL_library_init();
		OpenSSL_add_ssl_algorithms();
		SSL_load_error_strings();
		iSslMethod = SSLv23_client_method();
		iSslCtx = SSL_CTX_new(iSslMethod);
	}

	SSL_CTX* socket_manager::ssl_ctx()
	{
		if (!iSslStarted)
			start_ssl();
		return iSslCtx; 
	}
	#endif //NO_SSL

	bool socket_manager::do_select(fd_set& aReadFds, fd_set& aWriteFds, fd_set& aExceptFds)
	{
		timeval timeout = {0, 1000 };
		unlock();
		int result = ::select(0, &aReadFds, &aWriteFds, &aExceptFds, &timeout);
		lock();
		if (result == SOCKET_ERROR)
		{
			get_error();
			if (iError != WSAENOTSOCK)
				return false;
			else
			{
				// one of the sockets has been closed by another thread perhaps?
				clear_error();
				return true;
			}
		}

		for (fd_list::iterator i = iReadFds.begin(); i != iReadFds.end(); ++i)
		{
			if (i->second.first)
			{
				socket& theSocket = *i->first;
				if (theSocket.zombie())
					continue;
				if (FD_ISSET(theSocket.handle(), &aReadFds))
					theSocket.set_status(socket::StatusRead);
			}
		}
		for (fd_list::iterator i = iWriteFds.begin(); i != iWriteFds.end(); ++i)
		{
			if (i->second.first)
			{
				socket& theSocket = *i->first;
				if (theSocket.zombie())
					continue;
				if (FD_ISSET(theSocket.handle(), &aWriteFds))
					theSocket.set_status(socket::StatusWrite);
			}
		}
		for (fd_list::iterator i = iExceptFds.begin(); i != iExceptFds.end(); ++i)
		{
			if (i->second.first)
			{
				socket& theSocket = *i->first;
				if (theSocket.zombie())
					continue;
				if (FD_ISSET(theSocket.handle(), &aExceptFds))
					theSocket.set_status(socket::StatusExcept);
			}
		}

		return true;
	}

	void socket_manager::do_io()
	{
		bool needsProcessing = false;
		for (socket_list::iterator i = iSockets.begin(); i != iSockets.end();)
		{
			socket& theSocket = **i++;
			if (!theSocket.zombie() && theSocket.get_status() & socket::StatusWrite)
			{
				if (!theSocket.connected())
				{
					needsProcessing = true;
					theSocket.set_connected();
					add_notification(socket_manager_observer::NotifySocketConnected, theSocket);
				}
			}
		}
		for (socket_list::iterator i = iSockets.begin(); i != iSockets.end();)
		{
			socket& theSocket = **i++;
			if (!theSocket.zombie() && theSocket.get_status() & socket::StatusRead)
			{
				needsProcessing = true;
				theSocket.handle_input();
			}
		}
		for (socket_list::iterator i = iSockets.begin(); i != iSockets.end();)
		{
			socket& theSocket = **i++;
			if (!theSocket.zombie() && theSocket.get_status() & socket::StatusWrite)
			{
				needsProcessing = true;
				theSocket.handle_output();
			}
		}
		for (socket_list::iterator i = iSockets.begin(); i != iSockets.end();)
		{
			socket& theSocket = **i++;
			if (!theSocket.zombie() && theSocket.get_status() & socket::StatusExcept)
			{
				needsProcessing = true;
				theSocket.handle_error();
			}
		}
		for (socket_list::iterator i = iSockets.begin(); i != iSockets.end();)
		{
			socket& theSocket = **i++;
			theSocket.clear_status(static_cast<socket::status>(socket::StatusRead|socket::StatusWrite|socket::StatusExcept));
		}

		if (needsProcessing)
			iProcessor.task();
	}

	void socket_manager::add_notification(socket_manager_observer::notify_type aNotificationType, socket& aSocket, const notification_data& aData)
	{
		lock();
		notification_entry theEntry = { aNotificationType, aSocket, aData };
		iNotificationEntries.push_back(theEntry);
		unlock();
	}

	void socket_manager::process_notifications()
	{
		for (notification_entries::iterator i = iNotificationEntries.begin(); i != iNotificationEntries.end(); )
		{
			switch(i->iType)
			{
			case socket_manager_observer::NotifySocketAdded:
				notify_observers(i->iType, i->iSocket, static_cast<bool>(i->iData));
				break;
			case socket_manager_observer::NotifySocketConnected:
				notify_observers(i->iType, i->iSocket);
				break;
			case socket_manager_observer::NotifyData:
				notify_observers(i->iType, i->iSocket, static_cast<packet&>(i->iData));
				break;
			case socket_manager_observer::NotifyUnderflow:
				notify_observers(i->iType, i->iSocket);
				break;
			}
			iNotificationEntries.erase(i++);
			if (yield())
				return;
		}
	}

	void socket_manager::process_zombies()
	{
		for (socket_list::iterator i = iSockets.begin(); i != iSockets.end();)
		{
			if ((*i)->zombie())
				remove_socket(*i++);
			else
				++i;
		}
	}

	void socket_manager::process_resolved()
	{
		while(iResolver.is_next_ready())
		{
			resolver::entry resolverEntry(iResolver);
			iResolver.next_ready(resolverEntry);
			socket* theSocket = const_cast<socket*>(resolverEntry.iSocket);
			if (theSocket != 0)
			{
				theSocket->iAddress = resolverEntry.iAddress;
				if (theSocket->address() != INADDR_NONE)
				{
					if (!theSocket->connect())
					{
						get_error();
						remove_socket(theSocket);
					}
					else
					{
						if (theSocket->connected())
							notify_observers(socket_manager_observer::NotifySocketConnected, *theSocket);
					}
				}
				else
				{
					theSocket->set_error(resolverEntry.iError);
					remove_socket(theSocket);
				}
			}
			else
				notify_observers(socket_manager_observer::NotifyHostResolved, std::make_pair(resolverEntry.iHostName, resolverEntry.iHostNumber), resolverEntry.iAddress);
			if (resolverEntry.iRequester)
				resolverEntry.iRequester->host_resolved(resolverEntry.iHostName, resolverEntry.iHostNumber, resolverEntry.iAddress);
		}
	}

	void socket_manager::get_error()
	{
		lock();
		iError = ::WSAGetLastError();
		iErrorString = "";
		for (size_t i = 0; i < sizeof(sSocketErrors) / sizeof(sSocketErrors[0]); ++i)
		{
			if (sSocketErrors[i].iErrorCode == iError)
			{
				iErrorString = sSocketErrors[i].iErrorString;
				break;
			}
		}
		unlock();
	}

	void socket_manager::clear_error()
	{
		lock();
		iError = 0;
		iErrorString = "";
		unlock();
	}

	u_long socket_manager::to_address(const std::string& aHostName)
	{
		u_long address = inet_addr(aHostName.c_str());
		if (address != INADDR_NONE)
			return ntohl(address);

		if (iBlocking == socket::NonBlocking)
			return INADDR_NONE;

		hostent *he = gethostbyname(aHostName.c_str());
		if (he == NULL)
		{
			get_error();
			return INADDR_NONE;
		}
		
		sockaddr_in sa;
		memcpy(&sa.sin_addr, he->h_addr, he->h_length);
		return ntohl(sa.sin_addr.s_addr);
	}

	void socket_manager::add_socket(socket* aSocket, bool aListener)
	{
		lock();
		iSockets.push_back(aSocket);
		if (!in_task_thread())
			notify_observers(socket_manager_observer::NotifySocketAdded, *aSocket, aListener);
		else
			add_notification(socket_manager_observer::NotifySocketAdded, *aSocket, aListener);
		if (aSocket->connected())
		{
			if (!in_task_thread())
				notify_observers(socket_manager_observer::NotifySocketConnected, *aSocket);
			else
				add_notification(socket_manager_observer::NotifySocketConnected, *aSocket);
		}
		unlock();
	}

	void socket_manager::remove_socket(socket* aSocket)
	{
		lock();
		iSockets.remove(aSocket);
		iResolver.remove(*aSocket);
		fd_list::iterator fdi = iReadFds.find(aSocket);
		if (fdi != iReadFds.end())
			iReadFds.erase(fdi);
		fdi = iWriteFds.find(aSocket);
		if (fdi != iWriteFds.end())
			iWriteFds.erase(fdi);
		fdi = iExceptFds.find(aSocket);
		if (fdi != iExceptFds.end())
			iExceptFds.erase(fdi);
		for (notification_entries::iterator i = iNotificationEntries.begin(); i != iNotificationEntries.end(); )
			if (&i->iSocket == aSocket)
				iNotificationEntries.erase(i++);
			else
				++i;
		notify_observers(socket_manager_observer::NotifySocketRemoved, *aSocket);
		delete aSocket;
		unlock();
	}

	void socket_manager::notify_observer(socket_manager_observer& aObserver, socket_manager_observer::notify_type aType, const void* aParameter, const void* aParameter2)
	{
		socket& theSocket = *const_cast<socket*>(reinterpret_cast<const socket*>(aParameter));
		switch(aType)
		{
		case socket_manager_observer::NotifySocketAdded:
			aObserver.socket_added(theSocket, *reinterpret_cast<const bool*>(aParameter2));
			break;
		case socket_manager_observer::NotifySocketRemoved:
			aObserver.socket_removed(theSocket);
			break;
		case socket_manager_observer::NotifySocketConnected:
			aObserver.socket_connected(theSocket);
			break;
		case socket_manager_observer::NotifyData:
			aObserver.socket_data(theSocket, *reinterpret_cast<const packet*>(aParameter2));
			break;
		case socket_manager_observer::NotifyUnderflow:
			aObserver.socket_underflow(theSocket);
			break;		
		case socket_manager_observer::NotifyHostResolved:
			aObserver.host_resolved(reinterpret_cast<const std::pair<std::string, std::string>*>(aParameter)->first,
				reinterpret_cast<const std::pair<std::string, std::string>*>(aParameter)->second, 
				*reinterpret_cast<const u_long*>(aParameter2));
			break;
		}
	}
} // namespace lib