// serialization.h - v1.0.3
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#if !defined(LIB_SERIALIZATION)
#define LIB_SERIALIZATION

#include <climits>

// WARNING: This serialization library does not yet support endianness

namespace lib
{
	typedef unsigned char uint8_t;
	typedef unsigned int uint32_t;
	const unsigned int octet_bits = sizeof(uint8_t) * CHAR_BIT;

	class instream
	{
	public:
		virtual uint8_t read8() = 0;
		virtual uint32_t read32() = 0;
#ifdef LIB_SERIALIZATION_USE_STRING
		virtual void read_string(std::string& aString) = 0;
#endif
	};
	class outstream
	{
	public:
		virtual void write8(uint8_t a8) = 0;
		virtual void write32(uint32_t a32) = 0;
#ifdef LIB_SERIALIZATION_USE_STRING
		virtual void write_string(const std::string& aString) = 0;
#endif
	};
	class serializable
	{
	public:
		virtual void load(instream& aLoader) = 0;
		virtual void save(outstream& aSaver) const = 0;
	};

	class ptr_instream : public instream
	{
	public:
		struct error {};
		ptr_instream(const void* aPtr, size_t aLength) : iPtr(aPtr), iLength(aLength) {}
	private:
		virtual uint8_t read8()
		{
			if (iLength < sizeof(uint8_t))
				throw error();
			uint8_t ret = *static_cast<const uint8_t*>(iPtr);
			iPtr = static_cast<const uint8_t*>(iPtr) + 1;
			iLength -= sizeof(uint8_t);
			return ret;
		}
		virtual uint32_t read32()
		{
			if (iLength < sizeof(uint32_t))
				throw error();
			uint32_t ret = 0;
			for (size_t i = 0; i < sizeof(uint32_t); ++i)
				ret = (ret << octet_bits) | read8();
			return ret;
		}
#ifdef LIB_SERIALIZATION_USE_STRING
		virtual void read_string(std::string& aString)
		{
			size_t length = read32();
			if (iLength < length)
				throw error();
			aString.assign(static_cast<const char*>(iPtr), length);
			iPtr = static_cast<const char*>(iPtr) + length;
			iLength -= length;
		}
#endif
	private:
		const void* iPtr;
		size_t iLength;
	};
	class ptr_outstream : public outstream
	{
	public:
		struct error {};
		ptr_outstream(void* aPtr, size_t aLength) : iPtr(aPtr), iLength(aLength) {}
	private:
		virtual void write8(uint8_t a8)
		{
			if (iLength < sizeof(uint8_t))
				throw error();
			*static_cast<uint8_t*>(iPtr) = a8;
			iPtr = static_cast<uint8_t*>(iPtr) + 1;
			iLength -= sizeof(uint8_t);
		}
		virtual void write32(uint32_t a32)
		{
			if (iLength < sizeof(uint32_t))
				throw error();
			uint8_t bytes[sizeof(uint32_t)];
			for (size_t i = 0; i < sizeof(uint32_t); ++i)
			{
				bytes[(sizeof(uint32_t) - i) - 1] = a32 & UCHAR_MAX;
				a32 >>= octet_bits;
			}
			for (size_t i = 0; i < sizeof(uint32_t); ++i)
				write8(bytes[i]);
		}
#ifdef LIB_SERIALIZATION_USE_STRING
		virtual void write_string(const std::string& aString)
		{
			write32(aString.length());
			if (iLength < aString.length())
				throw error();
			memcpy(iPtr, aString.data(), aString.length());
			iPtr = static_cast<uint8_t*>(iPtr) + aString.length();
			iLength -= aString.length();
		}
#endif
	private:
		void* iPtr;
		size_t iLength;
	};
	class len_outstream : public outstream
	{
	public:
		len_outstream(size_t& aLength) : iLength(aLength) { iLength = 0; }
	private:
		virtual void write8(uint8_t /*a8*/)
		{
			iLength += sizeof(uint8_t);
		}
		virtual void write32(uint32_t /*a32*/)
		{
			iLength += sizeof(uint32_t);
		}
#ifdef LIB_SERIALIZATION_USE_STRING
		virtual void write_string(const std::string& aString)
		{
			write32(aString.length());
			iLength += aString.length();
		}
#endif
	private:
		len_outstream& operator=(const len_outstream&);
	private:
		size_t& iLength;
	};
}

#endif // LIB_SERIALIZATION