// threadable.cpp - v2.1.5
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "stdafx.h"
#include "threadable.h"

namespace lib
{
	threadable::threadable() : iState(Init), iThreadId(0), iBlockedCount(0)
	{
	}

	threadable::threadable(const threadable& aOther) : iState(Init), iThreadId(0), iBlockedCount(0)
	{
	}

	threadable& threadable::operator=(const threadable& aOther)
	{
		return *this;
	}

	threadable::~threadable()
	{
		abort();
	}

	void threadable::run()
	{
		iThreadHandle = ::CreateThread(NULL, 0, thread_proc, this, CREATE_SUSPENDED, &iThreadId);
		if (iThreadHandle)
		{
			iState = Running;
			if (::ResumeThread(iThreadHandle) == 0xFFFFFFFF)
			{
				iThreadHandle.close();
				iThreadId = 0;
				iState = Error;
			}
		}
		else
			iState = Error;
	}

	void threadable::cancel()
	{
		if (!in_task_thread())
		{
			abort();
			return;
		}
		bool canCancel = false;
		lock();
		if (is_running())
		{
			iState = Aborted;
			canCancel = true;
		}
		unlock();
		if (canCancel)
			throw cancellation();
	}

	void threadable::abort(bool aWait)
	{
		if (in_task_thread())
		{
			cancel();
			return;
		}
		handle waitHandle;
		lock();
		if (is_running())
		{
			::SuspendThread(iThreadHandle);
			if (!soft_abort() || blocked())
			{
				terminate();
				iState = Aborted;
				cleanup();
			}
			else
			{
				::ResumeThread(iThreadHandle);
				iState = Aborted;
				waitHandle = iThreadHandle;
			}
		}
		else if (is_aborted() && iThreadHandle)
		{
			waitHandle = iThreadHandle;
		}
		unlock();
		if (waitHandle && aWait)
			waitHandle.wait();
	}

	void threadable::block()
	{
		assert(in_task_thread());
		lock();
		++iBlockedCount;
		bool shouldCancel = is_aborted();
		unlock();
		if (shouldCancel)
			throw cancellation();
	}

	void threadable::unblock()
	{
		assert(in_task_thread());
		lock();
		--iBlockedCount;
		bool shouldCancel = is_aborted();
		unlock();
		if (shouldCancel)
			throw cancellation();
	}

	DWORD WINAPI threadable::thread_proc(LPVOID aThreadableObject)
	{
		threadable& threadableObject = *static_cast<threadable*>(aThreadableObject);
		try
		{
			threadableObject.task();
		}
		catch(const cancellation&)
		{
			// ok, quietly ignore
		}
		catch(...)
		{
			threadableObject.finish();
			throw;
		}
		threadableObject.finish();
		return 0;
	}

	bool threadable::duplicate_handle(handle& aHandle) const
	{
		lock();
		if (is_running())
			aHandle = iThreadHandle;
		unlock();
		return aHandle;
	}

	void threadable::finish()
	{
		lock();
		bool notify = in_task_thread();
		iThreadHandle.close();
		iThreadId = 0;
		iState = Finished;
		unlock();
		if (notify)
			finished();
	}

	const handle* threadable::wait(handle::timeout_interval aTimeoutInterval) const
	{
		handle waitHandle;
		if (!duplicate_handle(waitHandle))
			return &iThreadHandle;
		const handle* ret = waitHandle.wait(aTimeoutInterval);
		return ret == &waitHandle ? &iThreadHandle : ret;
	}

	const handle* threadable::msg_wait(handle::timeout_interval aTimeoutInterval) const
	{
		handle waitHandle;
		if (!duplicate_handle(waitHandle))
			return &iThreadHandle;
		const handle* ret = waitHandle.msg_wait(aTimeoutInterval);
		return ret == &waitHandle ? &iThreadHandle : ret;
	}

	const handle* threadable::wait_for_multiple(const handle& aHandle, handle::timeout_interval aTimeoutInterval) const
	{
		handle waitHandle;
		if (!duplicate_handle(waitHandle))
			return &iThreadHandle;
		const handle* ret = waitHandle.wait_for_multiple(aHandle, aTimeoutInterval);
		return ret == &waitHandle ? &iThreadHandle : ret;
	}

	const handle* threadable::msg_wait_for_multiple(const handle& aHandle, handle::timeout_interval aTimeoutInterval) const
	{
		handle waitHandle;
		if (!duplicate_handle(waitHandle))
			return &iThreadHandle;
		const handle* ret = waitHandle.msg_wait_for_multiple(aHandle, aTimeoutInterval);
		return ret == &waitHandle ? &iThreadHandle : ret;
	}

	const handle* threadable::wait_for_multiple(const handle_ref* aHandleRefs, std::size_t aCount, handle::timeout_interval aTimeoutInterval) const
	{
		handle waitHandle;
		if (!duplicate_handle(waitHandle))
			return &iThreadHandle;
		const handle* ret = waitHandle.wait_for_multiple(aHandleRefs, aCount, aTimeoutInterval);
		return ret == &waitHandle ? &iThreadHandle : ret;
	}

	const handle* threadable::msg_wait_for_multiple(const handle_ref* aHandleRefs, std::size_t aCount, handle::timeout_interval aTimeoutInterval) const
	{
		handle waitHandle;
		if (!duplicate_handle(waitHandle))
			return &iThreadHandle;
		const handle* ret = waitHandle.msg_wait_for_multiple(aHandleRefs, aCount, aTimeoutInterval);
		return ret == &waitHandle ? &iThreadHandle : ret;
	}

	void threadable::terminate()
	{
		lock();
		if (is_running())
		{
			::TerminateThread(iThreadHandle, 0);
			finish();
		}
		unlock();
	}

	bool threadable::soft_abort() const
	{
		return false;
	}
} // namespace lib