// threadable.h v2.1.4
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef LIB_THREADABLE
#define LIB_THREADABLE

#include "lockable.h"
#include "handle.h"

namespace lib
{
	class threadable : public lockable
	{
		// types
	private:
		enum state_e { Init, Running, Finished, Aborted, Error };
		struct cancellation {};
		// construction
	public:
		threadable();
		threadable(const threadable& aOther);
		threadable& operator=(const threadable& aOther);
		virtual ~threadable();
		// operations
	public:
		void run();
		void cancel();
		void abort(bool aWait = true);
		void block();
		void unblock();
		const handle* wait(handle::timeout_interval aTimeoutInterval = handle::TimeoutInfinite) const;
		const handle* msg_wait(handle::timeout_interval aTimeoutInterval = handle::TimeoutInfinite) const;
		const handle* wait_for_multiple(const handle& aHandle, handle::timeout_interval aTimeoutInterval = handle::TimeoutInfinite) const;
		const handle* msg_wait_for_multiple(const handle& aHandle, handle::timeout_interval aTimeoutInterval = handle::TimeoutInfinite) const;
		const handle* wait_for_multiple(const handle_ref* aHandleRefs, std::size_t aCount, handle::timeout_interval aTimeoutInterval = handle::TimeoutInfinite) const;
		const handle* msg_wait_for_multiple(const handle_ref* aHandleRefs = 0, std::size_t aCount = 0, handle::timeout_interval aTimeoutInterval = handle::TimeoutInfinite) const;
		bool is_running() const { return iState == Running; }
		bool is_finished() const { return iState == Finished; }
		bool is_aborted() const { return iState == Aborted; }
		bool error() const { return iState == Error; }
		handle thread_handle() const { return iThreadHandle; }
		DWORD thread_id() const { return iThreadId; }
	protected:
		bool in_task_thread() const { return ::GetCurrentThreadId() == iThreadId; }
		bool blocked() const { return iBlockedCount != 0; }
		// implementation
	private:
		static DWORD WINAPI thread_proc(LPVOID aThreadableObject);
		bool duplicate_handle(handle& aHandle) const;
		void finish();
		void terminate();
		virtual void task() = 0;
		virtual bool soft_abort() const;
		virtual void cleanup() {} // called after abort
		virtual void finished() {}
		// attributes
	private:
		volatile state_e iState;
		handle iThreadHandle;
		DWORD iThreadId;
		volatile std::size_t iBlockedCount;
	};
}

#endif //LIB_THREADABLE
