// timer.h
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef LIB_TIMER
#define LIB_TIMER

#include <list>
#include <vector>
#include "random.h"

namespace lib
{
	class waitable;

	class timer
	{
		// types
	public:
		typedef std::list<waitable*> waitable_list;

		// construction
	public:
		timer() {}
		virtual ~timer() {}

		// operations
	public:
		void add(waitable& aWaitable);
		void remove(waitable& aWaitable);
		bool has(const waitable& aWaitable) const;
		void tick();
		virtual unsigned long elapsed() = 0; // in ms

		// attributes
	private:
		waitable_list iWaitables;
		typedef std::vector<waitable_list::iterator> notifications;
		notifications iNotifications;
		random iRandom;
	};

	class waitable
	{
		// construction
	public:
		waitable(timer& aTimer, unsigned long aInterval) : iTimer(aTimer), iInterval(aInterval), iReady(false), iStartTime(iTimer.elapsed()) {}
		waitable(const waitable& aOther) : iTimer(aOther.iTimer), iInterval(aOther.iInterval), iReady(aOther.iReady), iStartTime(aOther.iStartTime) 
		{
			if (aOther.enabled())
				enable();
		}
		waitable& operator=(const waitable& aOther) 
		{ 
			iInterval = aOther.iInterval; 
			iReady = aOther.iReady; 
			iStartTime = aOther.iStartTime;
			if (aOther.enabled())
				enable();
			return *this;
		}
		virtual ~waitable() { disable(); }
		// operations
	public:
		bool enabled() const { return iTimer.has(*this); }
		void enable() { iTimer.add(*this); reset(); }
		void disable() { iTimer.remove(*this); }
		void tick() { if (!iReady && iTimer.elapsed() - iStartTime >= iInterval) { iReady = true; ready(); } }
		void reset() { iReady = false; iStartTime = iTimer.elapsed(); }
		void redo() { iReady = false; }
		unsigned long interval() const { return iInterval; }
		void set_interval(unsigned long aInterval) { iInterval = aInterval; }
		bool is_ready() const { return iReady; }

		// implementation
	private:
		virtual void ready() = 0;

		// attributes
	protected:
		timer& iTimer;
	private:
		unsigned long iInterval;
		bool iReady;
		unsigned long iStartTime;
	};
}

#endif //LIB_TIMER
