// vecarray.h - v2.4.1
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef LIB_VECARRAY
#define LIB_VECARRAY

#include <cassert>
#include <cstddef>
#include <stdexcept>
#include <memory>
#include <algorithm>
#include <iterator>
#include <type_traits>

namespace lib
{
	typedef std::tr1::true_type true_type;
	typedef std::tr1::false_type false_type;
	
	struct vecarray_overflow : public std::exception
	{
		const char* what() { return "vecarray overflow"; }
	};

	struct nocheck
	{
		inline static void test(bool aValid)
		{
			assert(aValid);
		}
	};

	template <typename Exception>
	struct check
	{
		inline static void test(bool aValid)
		{
			if (!aValid)
				throw Exception();
		}
	};

	namespace detail
	{
		class alignment_dummy;
		typedef void (*function_ptr)();
		typedef int (alignment_dummy::*member_ptr);
		typedef int (alignment_dummy::*member_function_ptr)();
		union max_align
		{
			char dummy0; short dummy1; int dummy2; long dummy3; long long dummy4;
			float dummy5; double dummy6; long double dummy7; 
			void* dummy8; 
			function_ptr dummy9; member_ptr dummy10; member_function_ptr dummy11;
		};

		template <typename T, typename IsScalar>
		struct vecarray_scalar_trait_impl
		{ 
			typedef T* pointer;
			typedef const T* const_pointer;
			typedef IsScalar type;
			typedef const T& parameter_type; 
			inline static void construct(pointer object, parameter_type value)
			{
				new (static_cast<void*>(object)) T(value);
			}
			inline static void destruct(pointer object)
			{
				object->~T();
			}
			inline static void destruct(pointer first, pointer last)
			{
				while(first != last)
				{
					first->~T();
					++first;
				}
			}
			inline static void uninitialized_copy(const_pointer first, const_pointer last, pointer dest)
			{
				std::uninitialized_copy(first, last, dest);
			}
			inline static void uninitialized_copy2(const_pointer first1, const_pointer last1, pointer dest1, const_pointer first2, const_pointer last2, pointer dest2)
			{
				std::uninitialized_copy(first1, last1, dest1);
				try
				{
					std::uninitialized_copy(first2, last2, dest2);
				}
				catch(...)
				{
					destruct(dest1, dest1 + (last1 - first1));
					throw;
				}
			}
		};
		template <typename T>
		struct vecarray_scalar_trait_impl<T, true_type>
		{ 
			typedef T* pointer;
			typedef const T* const_pointer;
			typedef true_type type;
			typedef T parameter_type; 
			inline static void construct(pointer object, parameter_type value)
			{
				*object = value;
			}
			inline static void destruct(pointer object)
			{
			}
			inline static void destruct(pointer first, pointer last)
			{
			}
			inline static void uninitialized_copy(const_pointer first, const_pointer last, pointer dest)
			{
				std::uninitialized_copy(first, last, dest);
			}
			inline static void uninitialized_copy2(const_pointer first1, const_pointer last1, pointer dest1, const_pointer first2, const_pointer last2, pointer dest2)
			{
				std::uninitialized_copy(first1, last1, dest1);
				std::uninitialized_copy(first2, last2, dest2);
			}
		};
	}

	template<typename T, std::size_t N, typename CheckPolicy = check<vecarray_overflow> >
	class vecarray
	{
	public:
		typedef T value_type;
		typedef T* pointer;
		typedef const T* const_pointer;
		typedef T& reference;
		typedef const T& const_reference;
		typedef pointer iterator;
		typedef const_pointer const_iterator;
		typedef std::reverse_iterator<iterator> reverse_iterator;
		typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
		typedef std::size_t size_type;
		typedef std::ptrdiff_t difference_type;
	private:
		typedef detail::vecarray_scalar_trait_impl<T, typename std::tr1::is_scalar<T>::type> scalar_trait;
	public:
		typedef typename scalar_trait::parameter_type parameter_type;
	
	private:
		union
		{
			detail::max_align dummy0;
			char iData[sizeof(T) * N];
		} iAlignedBuffer;
		size_type iSize;
	
	public:
		// construction
		vecarray() : iSize(0)
		{
		}
		vecarray(const vecarray& rhs) : iSize(0)
		{
			insert(begin(), rhs.begin(), rhs.end());
		}
		template <typename T2, std::size_t N2>
		vecarray(const vecarray<T2, N2>& rhs) : iSize(0)
		{
			insert(begin(), rhs.begin(), rhs.end());
		}
		vecarray(size_type n) : iSize(0)
		{
			insert(begin(), n, value_type());
		}
		vecarray(size_type n, parameter_type value) : iSize(0)
		{
			insert(begin(), n, value);
		}
		template<typename InputIterator>
		vecarray(InputIterator first, InputIterator last) : iSize(0)
		{
			insert(begin(), first, last);
		}
		~vecarray()
		{
			clear();
		}
		// traversals
		iterator begin() { return reinterpret_cast<pointer>(iAlignedBuffer.iData); }
		const_iterator begin() const { return reinterpret_cast<const_pointer>(iAlignedBuffer.iData); }
		iterator end() { return reinterpret_cast<pointer>(iAlignedBuffer.iData) + iSize; }
		const_iterator end() const { return reinterpret_cast<const_pointer>(iAlignedBuffer.iData) + iSize; }
		reverse_iterator rbegin() { return reverse_iterator(end()); }
		const_reverse_iterator rbegin() const { return const_reverse_iterator(end()); }
		reverse_iterator rend() { return reverse_iterator(begin()); }
		const_reverse_iterator rend() const { return const_reverse_iterator(begin()); }
		iterator locate(size_type position) { return reinterpret_cast<pointer>(iAlignedBuffer.iData) + position; }
		const_iterator locate(size_type position) const { return reinterpret_cast<const_pointer>(iAlignedBuffer.iData) + position; }
		bool empty() const { return iSize == 0; }
		bool full() const { return iSize == N; }
		size_type size() const { return iSize; }
		size_type available() const { return N - iSize; }
		size_type capacity() const { return N; }
		size_type max_size() const { return N; }
		size_type after(size_type position) const { return position < iSize ? iSize - position : 0; }
		// element access
		reference operator[](size_type n) { return *(begin() + n); }
		const_reference operator[](size_type n) const { return *(begin() + n); }
		reference at(size_type n) { if (n < iSize) return operator[](n); throw std::out_of_range("vecarray::at"); }
		const_reference at(size_type n) const { if (n < iSize) return operator[](n); throw std::out_of_range("vecarray::at"); }
		reference front() { return *begin(); }
		reference back() { return *(begin() + (iSize - 1)); }
		const_reference front() const { return *begin(); }
		const_reference back() const { return *(begin() + (iSize - 1)); }
		// modifiers
		vecarray& operator=(const vecarray& rhs)
		{
			if (&rhs != this)
				assign(rhs.begin(), rhs.end());
			return *this;
		}
		template<typename T2, std::size_t N2>
		vecarray& operator=(const vecarray<T2, N2>& rhs)
		{
			if (static_cast<const void*>(&rhs) != static_cast<const void*>(this))
				assign(rhs.begin(), rhs.end());
			return *this;
		}
		template <typename InputIterator>
		void assign(InputIterator first, InputIterator last)
		{
			clear();
			insert(begin(), first, last);
		}
		void assign(size_type n, parameter_type value)
		{
			clear();
			insert(begin(), n, value);
		}
		iterator insert(iterator position, parameter_type value)
		{
			insert(position, &value, &value+1);
			return position;
		}
		void insert(iterator position, size_type n, parameter_type value)
		{
			while(n > 0)
			{
				insert(position, value);
				++position;
				--n;
			}
		}
		template <typename InputIterator>
		void insert(iterator position, InputIterator first, InputIterator last)
		{
			insert_dispatch(position, first, last, typename std::tr1::is_integral<InputIterator>::type());
		}
		void insert(iterator position, iterator first, iterator last)
		{
			insert(position, static_cast<const_iterator>(first), static_cast<const_iterator>(last));
		}
		void insert(iterator position, const_iterator first, const_iterator last)
		{
			difference_type n = last - first;
			overflow_check(n);
			iterator theEnd = end();
			difference_type t = theEnd - position;
			if (t > 0)
			{
				if (t > n)
				{
					scalar_trait::uninitialized_copy(theEnd - n, theEnd, theEnd);
					iSize += n;
					std::copy_backward(position, theEnd - n, theEnd);
					std::copy(first, last, position);
				}
				else
				{
					scalar_trait::uninitialized_copy2(theEnd - t, theEnd, theEnd + (n - t), first + t, last, theEnd);
					iSize += n;
					std::copy(first, first + t, position);
				}
			}
			else
			{
				scalar_trait::uninitialized_copy(first, last, theEnd);
				iSize += n;
			}
		}
		iterator erase(iterator position) 
		{ 
			assert(iSize > 0);
			iterator last = end(); --last;
			std::copy(position + 1, end(), position);
			scalar_trait::destruct(last);
			--iSize;
			return position;
		}
		iterator erase(iterator first, iterator last) 
		{ 
			if (last != end())
				std::copy(last, end(), first);
			scalar_trait::destruct(end() - (last - first), end());
			iSize-=(last - first);
			return first;
		}
		void clear()
		{
			erase(begin(), end());
		}
		void push_back(parameter_type value)
		{
			push_back_dispatch(value, typename scalar_trait::type());
		}
		void pop_back()
		{
			erase(end() - 1);
		}
		void remove(parameter_type value, bool multiple = true)
		{
			for (iterator i = begin(); i != end(); )
			{
				if (*i == value)
				{
					erase(i);
					if (!multiple)
						return;
				}
				else
					++i;
			}
		}
		void resize(size_type n)
		{
			if (iSize > n)
				erase(begin() + n, end());
			else if (iSize < n)
				insert(end(), n - iSize, value_type());
		}
		void resize(size_type n, parameter_type value)
		{
			if (iSize > n)
				erase(begin() + n, end());
			else if (iSize < n)
				insert(end(), n - iSize, value);
		}
		template <typename T2, std::size_t N2>
		void swap(vecarray<T2, N2>& rhs)
		{
			vecarray tmp = rhs;
			rhs = *this;
			*this = tmp;
		}
		// equality
		template<typename T2, std::size_t N2>
		bool operator==(const vecarray<T2, N2>& rhs) const
		{
			return rhs.size() == size() && std::equal(begin(), end(), rhs.begin());
		}
		template<typename T2, std::size_t N2>
		bool operator!=(const vecarray<T2, N2>& rhs) const
		{
			return !operator==(rhs);
		}

	private:
		template <typename InputIterator>
		void insert_dispatch(iterator position, InputIterator first, InputIterator last, false_type)
		{
			while(first != last)
				insert(position++, *first++);
		}
		template <typename Integer>
		void insert_dispatch(iterator position, Integer n1, Integer n2, true_type)
		{
			size_type n = static_cast<size_type>(n1);
			value_type value = static_cast<value_type>(n2);
			while(n > 0)
			{
				insert(position, value);
				++position;
				--n;
			}
		}
		void push_back_dispatch(parameter_type value, false_type)
		{
			overflow_check(1);
			scalar_trait::construct(end(), value);
			++iSize;
		}
		void push_back_dispatch(parameter_type value, true_type)
		{
			overflow_check(1);
			*end() = value;
			++iSize;
		}
		void overflow_check(size_type n) const
		{
			CheckPolicy::test(iSize + n <= N);
		}
	};
}

#endif // LIB_VECARRAY
